{

  Chat codes WYSIWYG edition, v1.0

  This demo uses the same codes for chat formatting, but messages are
  typed in TRichViewEdit


  Codes:
  \b - bold on/off
  \i - italic on/off
  \u - underline on/off
  \cRRGGBB - text color
  \kRRGGBB - text background color
  \f"FONTNAME" - font name
  \sNN - font size

}

unit Unit1WYSIWYG;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, RVScroll, RichView, RVStyle, RVFuncs, RVItem,
  CRVData, CRVFData, ShellApi, Buttons, RVEdit, RVLinear, RVTypes,
  RVGrHandler, RVFileFuncs, RVStrFuncs, Math, Menus;

// Parameters for ApplyStyleConversion
const
  TEXT_BOLD       = 1;
  TEXT_ITALIC     = 2;
  TEXT_UNDERLINE  = 3;
  TEXT_APPLYFONTNAME  = 4;
  TEXT_APPLYFONTSIZE  = 5;
  TEXT_COLOR      = 6;
  TEXT_BACKCOLOR  = 7;


type

  TEmoticonCode = record
    Code: TRVUnicodeString;
    Index: Integer;
  end;

  TForm1 = class(TForm)
    rvs: TRVStyle;
    rv: TRichView;
    btnEnter: TButton;
    btnColor: TButton;
    btnBkColor: TButton;
    ColorDialog1: TColorDialog;
    btnArial: TButton;
    btnTimes: TButton;
    btnFontSize: TButton;
    rvs2: TRVStyle;
    rve: TRichViewEdit;
    btnBold: TSpeedButton;
    btnItalic: TSpeedButton;
    btnUnderline: TSpeedButton;
    btnEmoticon: TButton;
    pmEmoticon: TPopupMenu;
    Label1: TLabel;
    lblCode: TLabel;
    Edit1: TEdit;
    procedure btnBoldClick(Sender: TObject);
    procedure btnItalicClick(Sender: TObject);
    procedure btnUnderlineClick(Sender: TObject);
    procedure btnColorClick(Sender: TObject);
    procedure btnBkColorClick(Sender: TObject);
    procedure btnEnterClick(Sender: TObject);
    procedure btnArialClick(Sender: TObject);
    procedure btnTimesClick(Sender: TObject);
    procedure btnFontSizeClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure rvJump(Sender: TObject; id: Integer);
    procedure rveStyleConversion(Sender: TCustomRichViewEdit; StyleNo,
      UserData: Integer; AppliedToText: Boolean; var NewStyleNo: Integer);
    procedure rveCurTextStyleChanged(Sender: TObject);
    procedure btnEmoticonClick(Sender: TObject);
    procedure rveChange(Sender: TObject);
    procedure rveKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    EmoticonsGraphics: array of TGraphic;
    EmoticonsCodes: array of TEmoticonCode;
    FontName: String;
    FontSize: Integer;
    NoEmoticonDetection: Boolean;
    function GetStyleNo(CommandChar: Char; Color: TColor; StyleNo: Integer;
      const FontName: String; FontSize: Integer): Integer;
    function GetHypertextStyleNo(StyleNo: Integer): Integer;
    function Encode(rv: TCustomRichView): TRVUnicodeString;
    procedure ParseString(rv: TCustomRichView; s: TRVUnicodeString);
    procedure CreateEmoticons;
    procedure DestroyEmoticons;
    procedure FillEmoticonsMenu;
    function EmoticonAtIndex(const s: TRVUnicodeString; Index: Integer): Integer;
    procedure FindEmoticons(rve: TCustomRichViewEdit);
    procedure InsertEmoticon(rve: TCustomRichViewEdit; Index: Integer);
    procedure MenuItemInsertEmoticon(Sender: TObject);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{============================ EMOTICONS =======================================}
procedure TForm1.FormCreate(Sender: TObject);
begin
  CreateEmoticons;
  FillEmoticonsMenu;
  pmEmoticon.AutoHotkeys := maManual;
  // Setting mixed editing mode: do not allow changing paragraph style, but
  // allow changing text style
  rve.RVFOptions := rve.RVFOptions+[rvfoSaveTextStyles]-[rvfoSaveParaStyles];
  rve.RVFTextStylesReadMode := rvf_sInsertMerge;
  rve.RVFParaStylesReadMode := rvf_sInsertMap;
  rve.RTFReadProperties.TextStyleMode := rvrsAddIfNeeded;
  rve.RTFReadProperties.TextStyleMode := rvrsUseClosest;
  // ignoring undesirable content when pasting from RTF
  rve.RTFReadProperties.IgnorePictures := True;
  rve.RTFReadProperties.IgnoreNotes := True;
  rve.RTFReadProperties.IgnoreSequences := True;
  rve.RTFReadProperties.IgnoreTables := True;
  rve.RTFReadProperties.IgnoreBookmarks := True;
  // other settings
  rve.EditorOptions := rve.EditorOptions+[rvoDoNotWantReturns, rvoNoImageResize]-
    [rvoWantTabs];
  rve.VAlign := tlCenter;
  rve.Style.SpacesInTab := 4;
end;
{------------------------------------------------------------------------------}
procedure TForm1.FormDestroy(Sender: TObject);
begin
  DestroyEmoticons;
end;
{------------------------------------------------------------------------------}
// Initializing emoticons arrays
procedure TForm1.CreateEmoticons;
var i: Integer;
begin
  // EmoticonsGraphics array contains graphics for emoticons.
  // This code loads 0.bmp, 1.bmp, ..., 3.bmp as TBitmaps from Emoticons subfolder
  SetLength(EmoticonsGraphics, 4);
  for i := 0 to High(EmoticonsGraphics) do begin
    EmoticonsGraphics[i] := TBitmap.Create;
    EmoticonsGraphics[i].LoadFromFile(ExtractFilePath(Application.ExeName)+
      'Emoticons\'+IntToStr(i)+'.bmp');
  end;
  // Items in EmoticonsCodes array have the properties:
  // - Code - emoticon text
  // - Index - index in EmoticonsGraphics array
  SetLength(EmoticonsCodes, 8);
  EmoticonsCodes[0].Code := ':)';
  EmoticonsCodes[0].Index := 0;
  EmoticonsCodes[1].Code := ':-)';
  EmoticonsCodes[1].Index := 0;
  EmoticonsCodes[2].Code := ':(';
  EmoticonsCodes[2].Index := 1;
  EmoticonsCodes[3].Code := ':-(';
  EmoticonsCodes[3].Index := 1;
  EmoticonsCodes[4].Code := ':D';
  EmoticonsCodes[4].Index := 2;
  EmoticonsCodes[5].Code := ':-D';
  EmoticonsCodes[5].Index := 2;
  EmoticonsCodes[6].Code := ';)';
  EmoticonsCodes[6].Index := 3;
  EmoticonsCodes[7].Code := ';-)';
  EmoticonsCodes[7].Index := 3;
end;
{------------------------------------------------------------------------------}
// Destroying emoticons arrays
procedure TForm1.DestroyEmoticons;
var i: Integer;
begin
  for i := 0 to High(EmoticonsGraphics) do
    EmoticonsGraphics[i].Free;
  SetLength(EmoticonsGraphics, 0);
  SetLength(EmoticonsCodes, 0);
end;
{------------------------------------------------------------------------------}
// Returns index in EmoticonsCodes array, if s contains emoticon code starting
// from the character Index. Otherwise, returns -1.
function TForm1.EmoticonAtIndex(const s: TRVUnicodeString; Index: Integer): Integer;
var i: Integer;
begin
  for i := 0 to High(EmoticonsCodes) do
    if Copy(s, Index, Length(EmoticonsCodes[i].Code))=EmoticonsCodes[i].Code then
    begin
      Result := i;
      exit;
    end;
  Result := -1;
end;
{------------------------------------------------------------------------------}
// Inserts the Index-th emoticon in rve
procedure TForm1.InsertEmoticon(rve: TCustomRichViewEdit; Index: Integer);
var
  gr: TGraphic;
begin
  gr := RVGraphicHandler.CreateGraphic(
    TGraphicClass(EmoticonsGraphics[EmoticonsCodes[Index].Index].ClassType));
  gr.Assign(EmoticonsGraphics[EmoticonsCodes[Index].Index]);
  rve.BeginUndoGroup(rvutInsert);
  rve.SetUndoGroupMode(True);
  try
    if rve.InsertPicture('', gr, rvvaBaseline) then
      rve.SetCurrentItemExtraStrProperty(rvespAlt, EmoticonsCodes[Index].Code, True);
  finally
    rve.SetUndoGroupMode(False);
  end;
end;
{------------------------------------------------------------------------------}
// Change all codes to Emoticons in rve, as an editing operation; keeps the
// current selection
procedure TForm1.FindEmoticons(rve: TCustomRichViewEdit);
var i, SelStart, SelLen, SelStart2, SelLen2: Integer;
begin
  rve.BeginUpdate;
  RVGetSelection(rve, SelStart, SelLen);
  try
    for i := 0 to High(EmoticonsCodes) do
    begin
      rve.SetSelectionBounds(0, rve.GetOffsBeforeItem(0), 0, rve.GetOffsBeforeItem(0));
      while rve.SearchText(EmoticonsCodes[i].Code, [rvseoDown, rvseoMultiItem]) do
      begin
        RVGetSelection(rve, SelStart2, SelLen2);
        if SelStart2<min(SelStart, SelStart+SelLen) then
          dec(SelStart, Length(EmoticonsCodes[i].Code)-1)
        else if SelStart2<max(SelStart, SelStart+SelLen) then
          dec(SelLen, Length(EmoticonsCodes[i].Code)-1);
        InsertEmoticon(rve, i);
      end;
    end;
  finally
    RVSetSelection(rve, SelStart, SelLen);
    rve.EndUpdate;
  end;
end;
{============================ EDITING =========================================}
// TColor to RGB string
function GetColorStr(Color: TColor): String;
begin
  if Color=clNone then
    Color := clWhite;
  Result := IntToHex(ColorToRGB(Color),6);
  Result := Copy(Result,5,2)+Copy(Result,3,2)+Copy(Result,1,2);
end;
// RGB string to TColor
function GetStrColor(ColorValue: TRVUnicodeString): TColor;
var i,v: Integer;
begin
  ColorValue := Copy(ColorValue,5,2)+Copy(ColorValue,3,2)+Copy(ColorValue,1,2);
  Result := 0;
  for i := 1 to 6 do
  begin
    if RVCharInSet7(ColorValue[i], ['A'..'F']) then
      v := ord(ColorValue[i])-ord('A')+10
    else if RVCharInSet7(ColorValue[i], ['a'..'f']) then
      v := ord(ColorValue[i])-ord('a')+10
    else
      v := ord(ColorValue[i])-ord('0');
    Result := Result*16+v;
  end;
end;
{------------------------------------------------------------------------------}
// apply bold
procedure TForm1.btnBoldClick(Sender: TObject);
begin
  rve.ApplyStyleConversion(TEXT_BOLD);
end;
{------------------------------------------------------------------------------}
// apply italic
procedure TForm1.btnItalicClick(Sender: TObject);
begin
  rve.ApplyStyleConversion(TEXT_ITALIC);
end;
{------------------------------------------------------------------------------}
// apply underline
procedure TForm1.btnUnderlineClick(Sender: TObject);
begin
  rve.ApplyStyleConversion(TEXT_UNDERLINE);
end;
{------------------------------------------------------------------------------}
// apply text color
procedure TForm1.btnColorClick(Sender: TObject);
begin
  ColorDialog1.Color := rvs2.TextStyles[rve.CurTextStyleNo].Color;
  if ColorDialog1.Execute then
    rve.ApplyStyleConversion(TEXT_COLOR);
  rve.SetFocus;
end;
{------------------------------------------------------------------------------}
// apply text background color
procedure TForm1.btnBkColorClick(Sender: TObject);
begin
  if rvs2.TextStyles[rve.CurTextStyleNo].BackColor=clNone then
    ColorDialog1.Color := clWhite
  else
    ColorDialog1.Color := rvs2.TextStyles[rve.CurTextStyleNo].BackColor;
  if ColorDialog1.Execute then
    rve.ApplyStyleConversion(TEXT_BACKCOLOR);
  rve.SetFocus;
end;
{------------------------------------------------------------------------------}
// apply font Arial
procedure TForm1.btnArialClick(Sender: TObject);
begin
  FontName := 'Arial';
  rve.ApplyStyleConversion(TEXT_APPLYFONTNAME);
  rve.SetFocus;
end;
{------------------------------------------------------------------------------}
// apply font Times New Roman
procedure TForm1.btnTimesClick(Sender: TObject);
begin
  FontName := 'Times New Roman';
  rve.ApplyStyleConversion(TEXT_APPLYFONTNAME);
  rve.SetFocus;
end;
{------------------------------------------------------------------------------}
// apply font size
procedure TForm1.btnFontSizeClick(Sender: TObject);
var s: String;
begin
  s := IntToStr(rvs2.TextStyles[rve.CurTextStyleNo].Size);
  if not InputQuery('Enter Font Size', '&Size:', s) then
    exit;
  FontSize := StrToInt(s);
  rve.ApplyStyleConversion(TEXT_APPLYFONTSIZE);
  rve.SetFocus;
end;
{------------------------------------------------------------------------------}
// OnStyleConversion event: implementation of editing operations
procedure TForm1.rveStyleConversion(Sender: TCustomRichViewEdit; StyleNo,
  UserData: Integer; AppliedToText: Boolean; var NewStyleNo: Integer);
var FontInfo: TFontInfo;
begin
  FontInfo := TFontInfo.Create(nil);
  try
    FontInfo.Assign(rve.Style.TextStyles[StyleNo]);
    case UserData of
      TEXT_BOLD:
        if btnBold.Down then
          FontInfo.Style := FontInfo.Style+[fsBold]
        else
          FontInfo.Style := FontInfo.Style-[fsBold];
      TEXT_ITALIC:
        if btnItalic.Down then
          FontInfo.Style := FontInfo.Style+[fsItalic]
        else
          FontInfo.Style := FontInfo.Style-[fsItalic];
      TEXT_UNDERLINE:
        if btnUnderline.Down then
          FontInfo.Style := FontInfo.Style+[fsUnderline]
        else
          FontInfo.Style := FontInfo.Style-[fsUnderline];
      TEXT_APPLYFONTNAME:
        FontInfo.FontName := FontName;
      TEXT_APPLYFONTSIZE:
        FontInfo.Size     := FontSize;
      TEXT_COLOR:
        FontInfo.Color := ColorDialog1.Color;
      TEXT_BACKCOLOR:
        if ColorDialog1.Color=clWhite then
          FontInfo.BackColor := clNone
        else
          FontInfo.BackColor := ColorDialog1.Color;
    end;
    NewStyleNo := rve.Style.TextStyles.FindSuchStyle(StyleNo,FontInfo,RVAllFontInfoProperties);
    if NewStyleNo=-1 then begin
      rve.Style.TextStyles.Add;
      NewStyleNo := rve.Style.TextStyles.Count-1;
      rve.Style.TextStyles[NewStyleNo].Assign(FontInfo);
      rve.Style.TextStyles[NewStyleNo].Standard := False;
    end;
  finally
    FontInfo.Free;
  end;
end;
{------------------------------------------------------------------------------}
// updating editor UI
procedure TForm1.rveCurTextStyleChanged(Sender: TObject);
var fi: TFontInfo;
begin
  fi := rve.Style.TextStyles[rve.CurTextStyleNo];
  btnBold.Down      := fsBold      in fi.Style;
  btnItalic.Down    := fsItalic    in fi.Style;
  btnUnderline.Down := fsUnderline in fi.Style;
end;
{------------------------------------------------------------------------------}
// showing emoticons menu
procedure TForm1.btnEmoticonClick(Sender: TObject);
var pt: TPoint;
begin
  pt := Point(btnEmoticon.Left, btnEmoticon.Top+btnEmoticon.Height);
  pt := ClientToScreen(pt);
  pmEmoticon.Popup(pt.X, pt.Y);
end;
{------------------------------------------------------------------------------}
// creating emoticons menu
procedure TForm1.FillEmoticonsMenu;
var i: Integer;
    mi: TMenuItem;
begin
  // this creates menu items for each emoticons
  // possible improvements: adding only items with unique Emoticons[].Index,
  // because, for example, :) and :-) create the same image
  for i := 0 to High(EmoticonsCodes) do
  begin
    mi := TMenuItem.Create(Self);
    mi.Caption := EmoticonsCodes[i].Code;
    mi.Tag := i;
    mi.OnClick := MenuItemInsertEmoticon;
    pmEmoticon.Items.Add(mi);
  end;
end;
{------------------------------------------------------------------------------}
// inserting emoticon on menu click
procedure TForm1.MenuItemInsertEmoticon(Sender: TObject);
begin
  InsertEmoticon(rve, (Sender as TMenuItem).Tag);
end;
{------------------------------------------------------------------------------}
// rve.OnChange
procedure TForm1.rveChange(Sender: TObject);
begin
  // 1. detecting emoticons; unfortunately, this operation cannot be undone,
  // because undo calls OnChange again, and emoticons redetected
  if not NoEmoticonDetection then
  begin
    NoEmoticonDetection := True;
    try
      FindEmoticons(rve);
    finally
      NoEmoticonDetection := False;
    end;
  end;
  // 2. showing message to send
  lblCode.Caption := Encode(rve);
end;
{============================== CHAT FUNCTIONS ================================}
// Adding string from rve to rv, clearing rve
procedure TForm1.btnEnterClick(Sender: TObject);
begin
  //Edit1.Text := Encode(rve);
//  ParseString(rv, Encode(rve));
  ParseString(rv, Edit1.Text);
  rv.FormatTail;
  rv.Update;
  rv.DeleteUnusedStyles(True, True, True);

  rve.Clear;
  rve.DeleteUnusedStyles(True, True, True);
  rve.Format;
  rve.SetFocus;
  lblCode.Caption := '';
end;
{------------------------------------------------------------------------------}
// OnKeyPress event of rve
procedure TForm1.rveKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  // for any case, we check rve.InplaceEditor, and do nothing if we are
  // inside a table cell (because btnEnterClick will clear rve, so cell
  // inplace editor will be destroyed; but cell inplace editor must not
  // be destroyed in OnKeyDown)

  // the only way to insert table in this editor: copy it from another RichView
  if (Key=VK_RETURN) and (Shift=[]) and (rve.InplaceEditor=nil) then
  begin
    btnEnterClick(nil);
    Key := 0;
  end;
end;
{------------------------------------------------------------------------------}
// OnJump event of rv: opens URL in browser/e-mail client
procedure TForm1.rvJump(Sender: TObject; id: Integer);
var
  s: TRVUnicodeString;
  RVData: TCustomRVFormattedData;
  ItemNo: Integer;
begin
  TCustomRichView(Sender).GetJumpPointLocation(id, RVData, ItemNo);
  s := RVData.GetItemTextW(ItemNo);
  if not RVIsURL(s) and RVIsEmail(s) then
    s := 'mailto:'+s;
  ShellExecuteW(0, 'open', PRVUnicodeChar(s), nil, nil, SW_SHOW)
end;
{===================== CHAT FUNCTIONS: DECODING MESSAGE =======================} 
// Returning StyleNo basing on the command:
// 'b' - bold on/off;          - Color, FontName and FontSize are ignored
// 'i' - italic on/off;        - Color, FontName and FontSize are ignored
// 'u' - underline on/off;     - Color, FontName and FontSize are ignored
// 'c' - set text Color;       - FontName and FontSize are ignored
// 'k' - set background Color; - FontName and FontSize are ignored; clWhite works like clNone
// 'f' - font name;            - Color and FontSize are ignored
// 's' - font size;            - Color and FontName are ignored
function TForm1.GetStyleNo(CommandChar: Char; Color: TColor;
  StyleNo: Integer; const FontName: String; FontSize: Integer): Integer;
var fi: TFontInfo;
begin
  fi := TFontInfo.Create(nil);
  fi.Assign(rvs.TextStyles[StyleNo]);
  case CommandChar of
    'b':
      if fsBold in fi.Style then
        fi.Style := fi.Style-[fsBold]
      else
        fi.Style := fi.Style+[fsBold];
    'i':
      if fsItalic in fi.Style then
        fi.Style := fi.Style-[fsItalic]
      else
        fi.Style := fi.Style+[fsItalic];
    'u':
      if fsUnderline in fi.Style then
        fi.Style := fi.Style-[fsUnderline]
      else
        fi.Style := fi.Style+[fsUnderline];
    'c':
      fi.Color := Color;
    'k':
      if Color=clWhite then
        fi.BackColor := clNone
      else
        fi.BackColor := Color;
    'f':
      fi.FontName := FontName;
    's':
      fi.Size := FontSize;
  end;
  Result := rvs.FindTextStyle(fi);
  fi.Free;
end;
{------------------------------------------------------------------------------}
// Returns index of text styles having all properties of the StyleNo-th style,
// but hypertext, blue, highlighted and underlined
function TForm1.GetHypertextStyleNo(StyleNo: Integer): Integer;
var fi: TFontInfo;
begin
  fi := TFontInfo.Create(nil);
  fi.Assign(rvs.TextStyles[StyleNo]);
  fi.Jump := True;
  fi.Color := clBlue;
  fi.HoverColor := clWhite;
  fi.HoverBackColor := clBlue;
  fi.Style := fi.Style+[fsUnderline];
  Result := rvs.FindTextStyle(fi);
  fi.Free;
end;
{------------------------------------------------------------------------------}
// Returns position of the first delimiter character in pc, from 1 (or 0 is none)
function FindChar(pc: PRVUnicodeChar; Len: Integer; rv: TCustomRichView): Integer;
var i: Integer;
begin
  for i := 0 to Len-1 do
    if not RVCharInSet7(pc[i], ['.', ':', '/']) and rv.RVData.IsDelimiterW(pc[i]) then
    begin
      Result := i+1;
      exit;
    end;
  Result := 0;
end;
{------------------------------------------------------------------------------}
type TReadState = (rsNormal, rsFontNeedsFirstQuote, rsFontName, rsColor, rsFontSize);
// Adding string with codes
procedure TForm1.ParseString(rv: TCustomRichView; s: TRVUnicodeString);
var i: Integer;
    ParaNo, StyleNo, StartIndex, EmoticonIndex: Integer;
    CodeStarted: Boolean;
    ReadState: TReadState;
    ColorChar: Char;
    Value: TRVUnicodeString;
    {..................................}
    // replacing '\\' with '\'
    procedure RemoveDoubleBS(var str: TRVUnicodeString);
    var i: Integer;
        BSDeleted: Boolean;
    begin
      BSDeleted := False;
      for i := Length(str) downto 1 do
        if (str[i]='\') and (i>1) and (str[i-1]='\') and not BSDeleted then
        begin
          Delete(str, i, 1);
          BSDeleted := True;
        end
        else
          BSDeleted := False;
    end;
    {..................................}
    // Adds text from s[StartIndex] to s[EndIndex] in rv, with URL detection
    // Uses StyleNo and ParaNo. Changes ParaNo.
    procedure AddStringWithURLs(StartIndex, EndIndex: Integer);
    var str, CurrentWord, s1: TRVUnicodeString;
        p: Integer;
        pc, pcstart: PRVUnicodeChar;
        Len: Integer;
        HypertextStyleNo: Integer;
    begin
      if EndIndex<StartIndex then
        exit;
      HypertextStyleNo := -1;
      str := Copy(s, StartIndex, EndIndex-StartIndex+1);
      RemoveDoubleBS(str);
      pc := PRVUnicodeChar(str);
      pcstart := pc;
      Len := Length(str);
      while Len>0 do
      begin
        p := FindChar(pc, Len, rv);
        if p=1 then
        begin
          inc(pc);
          dec(Len);
          continue;
        end;
        if p=0 then
          p := Len+1;
        SetLength(CurrentWord, p-1);
        Move(pc^, PRVUnicodeChar(CurrentWord)^, (p-1)*sizeof(TRVUnicodeChar));
        if (RVIsURL(CurrentWord) or RVIsEmail(CurrentWord)) then
        begin
          if pcstart<pc then
          begin
            SetLength(s1, pc-pcstart);
            Move(pcstart^, PRVUnicodeChar(s1)^, (pc-pcstart)*sizeof(TRVUnicodeChar));
            rv.AddNLW(s1, StyleNo, ParaNo);
            ParaNo := -1;
          end;
          if HypertextStyleNo<0 then
            HypertextStyleNo := GetHypertextStyleNo(StyleNo);
          rv.AddNLW(CurrentWord, HypertextStyleNo, ParaNo);
          ParaNo := -1;
          inc(pc, p-1);
          dec(Len, p-1);
          pcstart := pc;
        end
        else
        begin
          inc(pc, p-1);
          dec(Len, p-1);
        end;
      end;
      if pcstart<pc then
      begin
        SetLength(s1, pc-pcstart);
        Move(pcstart^, PRVUnicodeChar(s1)^, (pc-pcstart)*sizeof(TRVUnicodeChar));
        rv.AddNLW(s1, StyleNo, ParaNo);
        ParaNo := -1;
      end;
    end;
    {..................................}
    // Adds the emoticon defined in EmoticonsCodes[Index]
    procedure AddEmoticon(Index: Integer);
    var gr: TGraphic;
    begin
      gr := RVGraphicHandler.CreateGraphic(
        TGraphicClass(EmoticonsGraphics[EmoticonsCodes[Index].Index].ClassType));
      gr.Assign(EmoticonsGraphics[EmoticonsCodes[Index].Index]);
      rv.AddPicture( '', gr, ParaNo, rvvaBaseline);
      rv.SetItemExtraStrProperty(rv.ItemCount-1, rvespAlt, EmoticonsCodes[Index].Code);
      ParaNo := -1;
    end;
    {..................................}
begin
  ParaNo := 0;
  CodeStarted := False;
  StyleNo := 0;
  StartIndex := 1;
  ColorChar := #0;
  ReadState := rsNormal;
  for i := 1 to Length(s) do
    if i>=StartIndex then
      case ReadState of
        rsColor: // expecting color value, RRGGBB
          begin
            if RVCharInSet7(s[i], ['0'..'9', 'a'..'f', 'A'..'F']) then
            begin
              Value := Value + s[i];
              if Length(Value)=6 then
              begin
                AddStringWithURLs(StartIndex, i-2-6);
                StartIndex := i+1;
                StyleNo := GetStyleNo(ColorChar, GetStrColor(Value), StyleNo, '', 0);
                ColorChar := #0;
                ReadState := rsNormal;
              end;
              continue;
            end;
            ReadState := rsNormal;
          end;
        rsFontNeedsFirstQuote: // expecting opening quote for the font name
          begin
            if s[i]<>'"' then
              ReadState := rsNormal
            else
              ReadState := rsFontName;
          end;
        rsFontName: // expecting font name, till the closing quote
          begin
            if s[i]<>'"' then
              Value := Value+s[i]
            else
            begin
              AddStringWithURLs(StartIndex, i-2-2-Length(Value));
              ReadState := rsNormal;
              StyleNo := GetStyleNo('f', clNone, StyleNo, Value, 0);
              StartIndex := i+1;
            end;
          end;
        rsFontSize: // expecting font size (either 2 digits, or 1 digit then nondigit;
                    // if the latter, then if this nondigit is a space-char,
                    // it will not be added in text)
          begin
            if RVCharInSet7(s[i], ['0'..'9']) then
              Value := Value + s[i];
            if (Length(Value)=2) or (i=Length(s)) or not RVCharInSet7(s[i], ['0'..'9']) then
            begin
              ReadState := rsNormal;
              if Length(Value)>0 then
              begin
                if RVCharInSet7(s[i], ['0'..'9']) then
                  AddStringWithURLs(StartIndex, i-2-Length(Value))
                else
                  AddStringWithURLs(StartIndex, i-2-(Length(Value)+1));
                if (Length(Value)=1) and (s[i]<>' ') and (i<>Length(s)) then
                  StartIndex := i
                else
                  StartIndex := i+1;
                StyleNo := GetStyleNo('s', clNone, StyleNo, '', StrToInt(Value));
              end;
              continue;
            end;
          end;
        rsNormal: // normal text, or beginning of fields
          case s[i] of
            '\':
              CodeStarted := not CodeStarted;
            'b', 'u', 'i', 'B', 'U', 'I':
              if CodeStarted then
              begin
                AddStringWithURLs(StartIndex, i-2);
                StartIndex := i+1;
                StyleNo := GetStyleNo(LowerCase(s[i])[1], clNone, StyleNo, '', 0);
                CodeStarted := False;
              end;
            'c', 'C', 'k', 'K':
              if CodeStarted then
              begin
                ColorChar := LowerCase(s[i])[1];
                Value := '';
                ReadState := rsColor;
                CodeStarted := False;
              end;
            'f', 'F':
              if CodeStarted then
              begin
                Value := '';
                ReadState := rsFontNeedsFirstQuote;
                CodeStarted := False;
              end;
            's', 'S':
              if CodeStarted then
              begin
                Value := '';
                ReadState := rsFontSize;
                CodeStarted := False;
              end;
            else
              begin
                CodeStarted := False;
                // emoticons
                EmoticonIndex := EmoticonAtIndex(s, i);
                if EmoticonIndex>=0 then
                begin
                  AddStringWithURLs(StartIndex, i-1);
                  AddEmoticon(EmoticonIndex);
                  StartIndex := i+Length(EmoticonsCodes[EmoticonIndex].Code);
                end;
              end;
          end;
      end;
  AddStringWithURLs(StartIndex, Length(s));
end;
{===================== CHAT FUNCTIONS: ENCODING MESSAGE =======================}
function TForm1.Encode(rv: TCustomRichView): TRVUnicodeString;
var
  i: Integer;
  TextStyle, CurTextStyle: TFontInfo;
  s: TRVUnicodeString;
begin
  Result := '';
  TextStyle := rve.Style.TextStyles[0];
  for i := 0 to rve.ItemCount-1 do
  begin
    if (i>0) and rve.IsFromNewLine(i) then
      Result := Result +' | ';
    if rve.GetItemStyle(i)>=0 then
    begin
      CurTextStyle := rve.Style.TextStyles[rve.GetItemStyle(i)];
      if (fsBold in CurTextStyle.Style)<>(fsBold in TextStyle.Style) then
        Result := Result+'\b';
      if (fsItalic in CurTextStyle.Style)<>(fsItalic in TextStyle.Style) then
        Result := Result+'\i';
      if (fsUnderline in CurTextStyle.Style)<>(fsUnderline in TextStyle.Style) then
        Result := Result+'\u';
      if GetColorStr(CurTextStyle.Color)<>GetColorStr(TextStyle.Color) then
        Result := Result+'\c'+GetColorStr(CurTextStyle.Color);
      if GetColorStr(CurTextStyle.BackColor)<>GetColorStr(TextStyle.BackColor) then
        Result := Result+'\k'+GetColorStr(CurTextStyle.BackColor);
      if AnsiCompareText(CurTextStyle.FontName, TextStyle.FontName)<>0 then
        Result := Result+'\f"'+CurTextStyle.FontName+'"';
      if CurTextStyle.Size<>TextStyle.Size then
      begin
        if CurTextStyle.Size<10 then
          Result := Result+'\s0'+IntToStr(CurTextStyle.Size)
        else if CurTextStyle.Size<50 then
          Result := Result+'\s'+IntToStr(CurTextStyle.Size)
        else
          Result := Result+'\s50';
      end;
      Result := Result+rve.GetItemText(i);
      TextStyle := CurTextStyle;
    end
    else if rve.GetItemStyle(i)=rvsPicture then
    begin
      rve.GetItemExtraStrProperty(i, rvespAlt, s);
      Result := Result+s;
    end;
  end;
end;

end.
